#include "Structures.H"
#include "List.H"
#include "List.C"
#include "FishUtils.H"
#include <math.h>

Fish::Fish()
  : px(0.0), py(0.0), vx(0.0), vy(0.0), fx(0.), fy(0.0), potential(0.0),
    myId(-1), grid(NULL)
{
}

Fish::Fish(double tpx, double tpy, double tvx, double tvy, int tid)
  : px(tpx), py(tpy), vx(tvx), vy(tvy), fx(0.), fy(0.0), potential(0.0),
    myId(tid), grid(NULL)
{
}

Fish *Fish::Clone()
{
  Fish *result = new Fish(px, py, vx, vy, myId);
  result->fx = fx;
  result->fy = fy;
  result->potential = potential;
  result->grid = grid;

  return result;
}

void Fish::TestPrint(ostream &os)
{
  os << "Fish #" << myId << ": (" << px << ", " << py << ") (" << vx << ", " <<
    vy << ")\n";
}

inline void Fish::CalcForce(Fish *other)
{
  double deltaX, deltaY, sqrDist;
  double dist, x, gravBase;
  
  /* okay, temp is a neighboring fish */
  deltaX = other->px - px;
  deltaY =  other->py - py;
  sqrDist = MAX( (SQR(deltaX) + SQR(deltaY)), SQR(gDxMin));
  
  if (sqrDist >= gSqrCutoff)
    return;
  else
    {
      dist = sqrt(sqrDist);
      if (sqrDist < .001111111)
        gravBase = .2 - 1.0/(300.0 * dist);
      else
        {
          x = 30.0 * dist;
          gravBase = .1/(x * x * x * x *x);
        }
    }
  
  fx += deltaX * gravBase;
  fy += deltaY * gravBase;
}

inline void Fish::CalcPotential(Fish *other)
{
  double deltaX, deltaY, sqrDist, dist;
  double pot;
  const double C = 0.1;
  const double rc = 1.0/30.0;
  
  deltaX = other->px - px;
  deltaY =  other->py - py;
  sqrDist = (SQR(deltaX) + SQR(deltaY));
  
  if (sqrDist >= gSqrCutoff)
    return;
  else if (sqrDist == 0.0)
    return;
  else
    {
      dist = sqrt(sqrDist);
      if (sqrDist < .001111111)
        pot = C*(2.0*dist - rc*(9.0/4.0 - log(rc) + log(dist)));
      else
        pot = -C*pow(rc, 5)/(4*pow(dist,4));
    }

  potential += pot;
}

inline double Fish::KineticEnergy()
{
  return (.5 * (vx * vx + vy * vy));
}


inline double Fish::UpdateFish(double deltaTime, Fish *moving)
{
  double sqrVel;
  
  moving->px = px + vx * deltaTime;
  moving->py = py + vy * deltaTime;
  moving->vx = vx + fx * deltaTime;
  moving->vy = vy + fy * deltaTime;
  sqrVel = vx * vx + vy * vy;

  /* first, check for bounces */
  if (moving->px < 0.0)
    {
      moving->px = -moving->px;
      moving->vx = -moving->vx;
    }
  else if (moving->px > gDomainSize)
    {
      moving->px = 2.0 * gDomainSize - moving->px;
      moving->vx = -moving->vx;
    }
      

  if (moving->py < 0.0)
    {
      moving->py = -moving->py;
      moving->vy = -moving->vy;
    }
  else if (moving->py > gDomainSize)
    {
      moving->py = 2.0 * gDomainSize - moving->py;
      moving->vy = -moving->vy;
    }

  return sqrVel;
}
     

GridElement::GridElement(double l, double r, double t, double b,
                         int grid_id)
  : left(l), right(r), top(t), bottom(b), isEdge(0), isGhost(0),
    myNodeNum(-1), myNeighbors(NULL), myId(grid_id), myCurrentFish(NULL),
    myNewFish(NULL)
  
{
}

int GridElement::IsNeighboring(int grid_id)
{
  for (int i = 0; i < 8; i++)
    if (myNeighbors[i]->myId == grid_id)
      return 1;

  return 0;
}

void GridElement::SwitchToNew()
{
  FishList *temp = myCurrentFish;
  myCurrentFish = myNewFish;
  myNewFish = temp;
}

void GridElement::TestPrint(ostream &os)
{
  os << "GE #" << myId << "(" << left << ", " << right << ", " << top <<
    ", " << bottom << ") " << myNodeNum << "\n";
}

void GridElement::QuickPrint(ostream &os)
{
  int row = myId / gGridWidth - 1;
  int col = myId % gGridWidth - 1;

  os << "(" << col << ", " << row << ") ";
}
               
  
void GridElement::AddNewFish(Fish *f)
{
  myNewFish->Add(f);
  f->grid = this;
}

void GridElement::RemoveNewFish(Fish *f)
{
  f->grid = NULL;
  myNewFish->Remove(f);
}
  
void GridElement::AddCurrentFish(Fish *f)
{
  myCurrentFish->Add(f);
  f->grid = this;
}

void GridElement::RemoveCurrentFish(Fish *f)
{
  f->grid = NULL;
  myCurrentFish->Remove(f);
}


Node::Node(int node_id)
  : myElements(NULL), HisGhosts(NULL), MyGhosts(NULL),
    myElementsAndGhosts(NULL), myNumElementsAndGhosts(-1), outgoingFish(NULL),
    myId(node_id), allNodes(NULL), newFish(NULL), currentFish(NULL),
    numFish(-1), numNodes(-1), mainNeighbors(NULL), theGrid(NULL),
    fishInfoOutBuffer(NULL), fishInfoInBuffer(NULL), myFish(NULL),
    incomingBuffer(NULL), outgoingBuffer(NULL)
{
}

Node::Node(int node_id, int number_nodes, int number_fish, Fish **al)
  :  myElements(NULL), HisGhosts(NULL), MyGhosts(NULL),
     myElementsAndGhosts(NULL), myNumElementsAndGhosts(-1),
     outgoingFish(NULL), myId(node_id), allNodes(NULL), newFish(NULL),
     currentFish(al), numFish(number_fish), numNodes(number_nodes),
     mainNeighbors(NULL), theGrid(NULL), fishInfoOutBuffer(NULL),
     fishInfoInBuffer(NULL), myFish(NULL), incomingBuffer(NULL),
     outgoingBuffer(NULL)
{
  /* Geez, so much to do, so little time...  Things this routine has to do:
     o allocate all the nodes for the neighbors..
     o fill the nodes lists appropriately...
       all myElements lists are from one generator.
       all outgoingFish are same generator as the myFish from gridElements..
       all myGhosts from another generator..
       all HisGhosts are from separate generators...
       myFish is a completely separate Fish list...
     o allocate the grid, and initialize it right..
     */

  int i;
  GEListGenerator *myElementsGen, *myGhostsGen;
  FishListGenerator flg(al, number_fish);
  FishListGenerator temp_flg(al, number_fish);
  FishListGenerator another_flg(al, number_fish);
  FishListGenerator outgoing_flg(al, number_fish);
  
  Node *n;
  GridElement *ge;
  
  myFish = temp_flg.GenerateList();
  
  allNodes = new (Node*)[numNodes];

  for (i = 0; i < numNodes; i++)
    {
      if (i != myId)
        allNodes[i] = new Node(i);
      else
        allNodes[i] = this;
    }

  /* Create the secondary fish list */
  newFish = new (Fish *)[number_fish];
  for (i = 0; i < number_fish; i++)
    newFish[i] = al[i]->Clone();
  
  /* allocate the grid */
  theGrid = new Grid(allNodes, &flg, &another_flg, numNodes);
  
  myElementsGen = theGrid->CreateGenerator();
  myGhostsGen = theGrid->CreateGenerator();

  for (i = 0; i < numNodes; i++)
    {
      GEListGenerator *temp = theGrid->CreateGenerator();
      allNodes[i]->myElements = myElementsGen->GenerateList();
      allNodes[i]->MyGhosts = myGhostsGen->GenerateList();
      allNodes[i]->HisGhosts = temp->GenerateList();
      allNodes[i]->outgoingFish = outgoing_flg.GenerateList();
      delete temp;
    }

  delete myElementsGen;
  delete myGhostsGen;

  theGrid->PartitionGrid(numNodes);
  mainNeighbors = theGrid->ComputeNeighborInfo(myId);

  CreateBuffers();

  /* remember to create myElementsAndGhost */
  mainNeighbors->Reset();
  myNumElementsAndGhosts = myElements->NumElements();
  
  while ((n = mainNeighbors->Next()))
    myNumElementsAndGhosts += n->MyGhosts->NumElements();

  myElementsAndGhosts = new (GridElement *)[ myNumElementsAndGhosts];

  mainNeighbors->Reset();
  i = 0;
  myElements->Reset();

  while ((ge = myElements->Next()))
    myElementsAndGhosts[i++] = ge;
  
  while ((n = mainNeighbors->Next()))
    {
      n->MyGhosts->Reset();
      while ((ge = n->MyGhosts->Next())) 
        myElementsAndGhosts[i++] = ge;
    }
  

  /* Figure out what grid elements need to be calculated before the
     stages */
  ComputeGridDependencies();
  
  /* Deposit all the little fish in the right place */
  for (i = 0; i < numFish; i++)
    {
      ge = theGrid->GEFromPoint(currentFish[i]->px, currentFish[i]->py);
      if (ge->myNodeNum == myId)
        {
          ge->AddCurrentFish(currentFish[i]);
          myFish->Add(currentFish[i]);
        }
      else if (ge->isGhost)
        ge->AddCurrentFish(currentFish[i]);
    }

  /* fill in the fishInfoBuffers */
  if (myId == 0)
    {
      fishInfoInBuffer = new (FishData *)[numNodes];
      for (i = 0; i < numNodes; i++)
        fishInfoInBuffer[i] = (FishData *)
          malloc(sizeof(FishData) * (numFish +2));
    }
  else
    fishInfoOutBuffer = (FishData *) malloc(sizeof(FishData) * (numFish + 2));

}

void Node::TestPrint(ostream &os)
{
  os << "Node #" << myId << "\n"; 
}

void Node::FillDependencies(Grid *theGrid, GEList *myElements)
{
  myStage1Dependencies = theGrid->CreateGenerator()->GenerateList();
  myStage2Dependencies = theGrid->CreateGenerator()->GenerateList();
  GridElement *current, *testing;
  
  /* HisGhosts are the Stage1 dependencies, since these are the only
     grids owned by this node that touch the neighbors area... Hence,
     only stuff in those grids can possibly move into the neigbhors */

  for (HisGhosts->Reset(); ((current = HisGhosts->Next()));)
    myStage1Dependencies->Add(current);

  /* Now for Stage 2 stuff... These consist of all the myElements that
     touch stuff in HisGhosts */

  for (HisGhosts->Reset(); ((current = HisGhosts->Next()));)
    for (myElements->Reset(); ((testing = myElements->Next()));)
      {
        if ((testing->IsNeighboring(current->myId)) &&
            !(myStage2Dependencies->Member(testing)))
          myStage2Dependencies->Add(testing);          
      }       
}

void Node::ComputeGridDependencies()
{
  NodeList *myStageNodes = NodeListGenerator(allNodes, gNumberNodes).
    GenerateList();
  
  GridElement *ge;
  
  Node *current, *best;
  int place;

  myStage1Leftovers = theGrid->CreateGenerator()->GenerateList();
  myStage2Leftovers = theGrid->CreateGenerator()->GenerateList();
  myStage1Ordering = new (Node*)[mainNeighbors->NumElements()];
  myStage2Ordering = new (Node*)[mainNeighbors->NumElements()];
  
  for (myElements->Reset(); ((ge = myElements->Next()));)
    myStage1Leftovers->Add(ge);
  
  /* First, need to create and fill all the stage dependency lists */
  for (mainNeighbors->Reset(); ((current = mainNeighbors->Next()));)
    current->FillDependencies(theGrid, myElements);

  /* Second, put all the neighboring nodes in the stage lists */
  for (mainNeighbors->Reset(); ((current = mainNeighbors->Next()));)
    myStageNodes->Add(current);

  /* Now, do stage 1...  The criteria I use is to choose the neighbor
     with the most grids to send in Stage 1 (which happens to be
     given by MyGhosts), */

  place = 0;
  while (myStageNodes->NumElements())
    {
      best = NULL;
      /* okay, pick guy with the most MyGhosts */
      for (myStageNodes->Reset(); ((current = myStageNodes->Next()));)
        if (best == NULL)
          best = current;
        else if (best->MyGhosts->NumElements() <
                 current->MyGhosts->NumElements())
          best = current;

      /* Okay, take all the dependences that are in the best, and remove
         them from all guys */

      for (best->myStage1Dependencies->Reset();
           ((ge = best->myStage1Dependencies->Next()));)
        {
          for (mainNeighbors->Reset(); ((current = mainNeighbors->Next()));)
            {
              if (current->myStage2Dependencies->Member(ge))
                current->myStage2Dependencies->Remove(ge);
              if (current == best)
                continue;
              if (current->myStage1Dependencies->Member(ge))
                current->myStage1Dependencies->Remove(ge);
            }
          if (myStage1Leftovers->Member(ge))
            myStage1Leftovers->Remove(ge);
        }

      myStage1Ordering[place++] = best;
      myStageNodes->Remove(best);

      /* it could be that some of the other dependencies have zero
         left now... we can seen them right away too, so let's
         check */

      for (myStageNodes->Reset(); ((current = myStageNodes->Next()));)
        if (current->myStage1Dependencies->NumElements() == 0)
          {
            myStage1Ordering[place++] = current;
            myStageNodes->Remove(current);
          }
    }

  /* Now, do the same thing for stage two.. This time we key off of
     HisGhosts */

  for (mainNeighbors->Reset(); ((current = mainNeighbors->Next()));)
    myStageNodes->Add(current);

  place = 0;
  while (myStageNodes->NumElements())
    {
      best = NULL;
      /* okay, pick guy with the most MyGhosts */
      for (myStageNodes->Reset(); ((current = myStageNodes->Next()));)
        if (best == NULL)
          best = current;
        else if (best->HisGhosts->NumElements() <
                 current->HisGhosts->NumElements())
          best = current;

      /* Okay, take all the dependences that are in the best, and remove
         them from all guys */

      for (best->myStage2Dependencies->Reset();
           ((ge = best->myStage2Dependencies->Next()));)
        {
          for (mainNeighbors->Reset(); ((current = mainNeighbors->Next()));)
            {
              if (current == best)
                continue;
              if (current->myStage2Dependencies->Member(ge))
                current->myStage2Dependencies->Remove(ge);
              
            }
          if (myStage1Leftovers->Member(ge))
            myStage1Leftovers->Remove(ge);
        }
      
      myStage2Ordering[place++] = best;
      myStageNodes->Remove(best);

      /* it could be that some of the other dependencies have zero
         left now... we can seen them right away too, so let's
         check */

      for (myStageNodes->Reset(); ((current = myStageNodes->Next()));)
        if (current->myStage2Dependencies->NumElements() == 0)
          {
            myStage2Ordering[place++] = current;
            myStageNodes->Remove(current);
          }
    }

#ifdef _DEBUG_
  cout << "Dependencies: " << myId << "\n";
  cout << "Stage One:\n";

  int total = 0;
  for (int x = 0; x < mainNeighbors->NumElements(); x++)
    {
      cout << "  NeighborNode " << myStage1Ordering[x]->myId << " has "
           << myStage1Ordering[x]->myStage1Dependencies->NumElements()
             << "\n";
      total += myStage1Ordering[x]->myStage1Dependencies->NumElements();
    }

  cout << "\nStage Two:\n";
  for (int x = 0; x < mainNeighbors->NumElements(); x++)
    {
      cout << "  NeighborNode " << myStage2Ordering[x]->myId << " has "
           << myStage2Ordering[x]->myStage2Dependencies->NumElements()
             << "\n";
      total += myStage2Ordering[x]->myStage2Dependencies->NumElements();
    }

  cout << "\nLeftover: " << myStage1Leftovers->NumElements() << "\n";
  total += myStage1Leftovers->NumElements();
  cout << "Total: " << total << "\n\n";
#endif

  /* Now slpit up the leftovers.. */

  int half = myStage1Leftovers->NumElements() / 4;
  GridElement *temp;

  myStage1Leftovers->Reset();
  for (int x = 0; x < half; x++)
    {
      temp = myStage1Leftovers->Next();
      myStage1Leftovers->Remove(temp);
      myStage2Leftovers->Add(temp);
    }
    

}

Grid::Grid(Node **node_list, FishListGenerator *fgen,
           FishListGenerator *curgen, int numberNodes)
  : myNodeList(node_list), myNumberNodes(numberNodes)
{
  /* we have to allocate enough grid elements to cover the grid area, plus
     a edge border of shadowing grid elements... */

  int total = gGridWidth * gGridWidth;
  int place = 0;
  GridElement *current;
  
  myGrid = new (GridElement*)[total];

  /* fill the grid area in row major order */
  for (double y = -gCutoff; y < gDomainSize + gCutoff; y+= gCutoff)
    for (double x = -gCutoff; x < gDomainSize + gCutoff; x+= gCutoff)
      {
        myGrid[place] = new GridElement(x, x + gCutoff, y, y + gCutoff,
                                        place);
        place++;
      }

  for (int row = 0; row < gGridWidth; row++)
    for (int col = 0; col < gGridWidth; col++)
      {

        current = myGrid[row * gGridWidth + col];

        if ((row == 0) || (col == 0) || (row == (gGridWidth -1)) ||
            (col == (gGridWidth - 1)))
          /* then we are on an edge, and should not fill in the
             neighbor list */
          current->isEdge = 1;
        else
          {
            current->myNeighbors = new (GridElement*)[8];
            
            place = 0;
            for (int neigh_row = -1; neigh_row <= 1; neigh_row++)
              for (int neigh_col = -1; neigh_col <= 1; neigh_col++)
                if (!((neigh_row == 0) && (neigh_col == 0)))
                  current->myNeighbors[place++] =
                    myGrid[ (row + neigh_row) * gGridWidth + col + neigh_col];
            
          }
        current->myNewFish = fgen->GenerateList();
        current->myCurrentFish = curgen->GenerateList();
      }
}

Grid::~Grid()
{
}

void Grid::PartitionGrid(int number)
{
  int region1[1] = { 1};
  int region2[1] = { 2};
  int region3[1] = { 3};
  int region4[2] = {2, 2};
  int region5[2] = { 3, 2};
  int region6[2] = { 3, 3};
  int region7[2] = { 4, 3};
  int region8[2] = { 4, 4};
  int region9[3] = { 3, 3, 3};
  int region10[3] = { 4, 3, 3};
  int region11[3] = { 4, 4, 3};
  int region12[3] = { 4, 4, 4};
  int region13[3] = { 5, 4, 4};
  int region14[3] = { 5, 5, 4};
  int region15[3] = { 5, 5, 5};
  int region16[4] = { 4, 4, 4, 4};
  
  switch (number)
    {
    case 1:
      ComputeRegions(1, region1);
      break;
    case 2:
      ComputeRegions(1, region2);
      break;
    case 3:
      ComputeRegions(1, region3);
      break;
    case 4:
      ComputeRegions(2, region4);
      break;
    case 5:
      ComputeRegions(2, region5);
      break;
    case 6:
      ComputeRegions(2, region6);
      break;
    case 7:
      ComputeRegions(2, region7);
      break;
    case 8:
      ComputeRegions(2, region8);
      break;
    case 9:
      ComputeRegions(3, region9);
      break;
    case 10:
      ComputeRegions(3, region10);
      break;
    case 11:
      ComputeRegions(3, region11);
      break;
    case 12:
      ComputeRegions(3, region12);
      break;
    case 13:
      ComputeRegions(3, region13);
      break;
    case 14:
      ComputeRegions(3, region14);
      break;
    case 15:
      ComputeRegions(3, region15);
      break;
    case 16:
      ComputeRegions(4, region16);
      break;
    }
}

void Grid::ComputeRegions(int rows, int *num_cols_in_row)
{
  int max_columns = 0;
  int number_columns = 0;
  int *row_starts;
  int *col_starts;
  int cols_so_far;
  int node;
  int row, col;
  
  for (row = 0; row < rows; row++)
    {
      max_columns = MAX(max_columns, num_cols_in_row[row]);
      number_columns += num_cols_in_row[row];
    }

  row_starts = new int[rows];
  col_starts = new int[max_columns];

  /* compute where each row should begin */

  cols_so_far = 0;
  for (row = 0; row < rows; row++)
    {
      if (row == 0)
        row_starts[row] = 0;
      else
        row_starts[row] = (int) ceil( ((double)cols_so_far)/
                                    ((double) number_columns)
                                    * ((double) (gGridWidth -2)));
      cols_so_far += num_cols_in_row[row];
    }

  node = 0;
  
  /* now, for each row, compute where its columns should start and end */
  for (row = 0; row < rows; row++)
    {
      int num_cols = num_cols_in_row[row];
      int row_end, col_end;

      for (col = 0; col < num_cols; col++)
        {
          if (col == 0)
            col_starts[0] = 0;
          else
            col_starts[col] = (int) ceil( ((double)col)/((double)num_cols)
                                          * ((double) (gGridWidth - 2)));
        }

      /* now that we have the column boundaries, register a node for
         each area */

      if (row != (rows-1))
        row_end = row_starts[row + 1];
      else
        row_end = gGridWidth - 2;

      for (col = 0; col < num_cols; col++)
        {
          if (col != (num_cols - 1))
            col_end = col_starts[col + 1];
          else
            col_end = gGridWidth - 2;

          SetNodeRegion(col_starts[col], row_starts[row], col_end, row_end,
                        node);

          node++;
        }
          
    }
  delete [] row_starts;
  delete [] col_starts;
}

NodeList *Grid::ComputeNeighborInfo(int myN)
{
  Node *myNode = myNodeList[myN];
  GridElement *current, *temp;
  Node *destNode;
  NodeListGenerator nlg(myNodeList, myNumberNodes);
  NodeList *result = nlg.GenerateList();
  
  /* for every grid element associated with this node, look at the
     other grid elements surrounding it... If it is owned by another
     node, then the ge this node owns will be a ghost for him, and the
     node we just looked at will be a ghost for us */
  myMainNode = myN;
  
  myNode->myElements->Reset();
  
  while ((current = (myNode->myElements->Next())))
    for (int i = 0; i < 8; i++)
      if ((current->myNeighbors[i]->myNodeNum != myN) &&
          !(current->myNeighbors[i]->isEdge))
        {
          temp = current->myNeighbors[i];
          temp->isGhost = 1;
          destNode = myNodeList[temp->myNodeNum];

          if (!(destNode->HisGhosts->Member(current)))
            destNode->HisGhosts->Add(current);
          if (!(destNode->MyGhosts->Member(temp)))
            destNode->MyGhosts->Add(temp);
        }

  for (int t = myNumberNodes - 1; t >= 0; t--)
    if ((t != myN) && (myNodeList[t]->HisGhosts->NumElements() != 0))
      result->Add(myNodeList[t]);

  return result;
}

GridElement *Grid::GEFromPoint(double x, double y)
{
  int col = (int) (x * gInvCutoff + 1);
  int row = (int) (y * gInvCutoff + 1);

  return myGrid[row * gGridWidth + col];
}

GEListGenerator *Grid::CreateGenerator()
{
  return new GEListGenerator(myGrid, gGridWidth * gGridWidth);
}

int Grid::VerifyLayout()
{
  /* Things we need to verify...
     o Using GEFromPoint, make sure all the grid elements are where they
     should be..
     o Make sure all the neighbors on each grid is right..
     */

  GridElement *temp, *temp2;
  int correct = 1;
  double y,x;
  
  for (y = -gCutoff/2; y < (gDomainSize + gCutoff); y+= gCutoff)
    for (x = -gCutoff/2; x < (gDomainSize + gCutoff); x+= gCutoff)
      {
        temp = GEFromPoint(x,y);

        if ((temp->left > x) || (temp->right < x) || (temp->top > y)
            || (temp->bottom < y))
          {
            correct = 0;
            cerr << "Invalid grid when checking (" << x << ", " << y <<
              "). Got: " << temp->myId << "\n";
          }
      }

  for (y = gCutoff/2; y < gDomainSize; y+= gCutoff)
    for (x = gCutoff/2; x < gDomainSize; x+= gCutoff)
      {
        int place = 0;
        temp = GEFromPoint(x,y);
        
        for (double minor_y = -gCutoff; minor_y <= gCutoff; minor_y += gCutoff)
          for (double minor_x = -gCutoff; minor_x <= gCutoff;
               minor_x += gCutoff)
            if (!((minor_y == 0.0) && (minor_x == 0.0)))
              {
                
                temp2 = GEFromPoint(x + minor_x, y + minor_y);
                if (temp->myNeighbors[place] != temp2)
                  {
                    correct = 0;
                    cerr << "Incorrect neighbor when checking point (" <<
                      x << ", " << y << ") at (" << minor_x << ", " <<
                      minor_y << ")\n";
                  }
                place++;
              }
      }

  /* make sure each grid element is claimed by a node, and that
     that node has it in its list */

  for (int row = 1; row < gGridWidth-1; row++)
    for (int col = 1; col < gGridWidth-1; col++)
      {
        temp = myGrid[row * gGridWidth + col];
        if (temp->myNodeNum < 0)
          {
            correct = 0;
            cerr << "Unclaimed node: ";
            temp->QuickPrint(cerr);
            cerr << "\n";
          }
        else if (!(myNodeList[temp->myNodeNum]->myElements->Member(temp)))
          {
            correct = 0;
            cerr << "Node does not have grid in list: ";
            temp->QuickPrint(cerr);
            cerr << "\n";
          } 
        
      }

  return correct;
  
}

void Node::GraphLayout(ostream &os)
{
  for (int row = 0; row < gGridWidth; row++)
    {
      for (int col = 0; col < gGridWidth; col++)
        {
          GridElement *current = theGrid->GridArray()[row * gGridWidth + col];
          
          if (current->isEdge)
            os << ".";
          else
            os << (char) ('A' + current->myNodeNum);
        }
      os << "\n";
    }
  os << "\n";
 
  for (int i = 0; i < numNodes; i++)
    {
      os << "  Node " << (char) ('A' + i) << " has " <<
        allNodes[i]->myElements->NumElements() << "\n";
    }
  os << "\n\n";
}

void Node::GraphNeighbors(ostream &os)
{
  for (int row = 0; row < gGridWidth; row++)
    {
      for (int col = 0; col < gGridWidth; col++)
        {
          GridElement *current = theGrid->GridArray()[row * gGridWidth + col];
          
          if (current->isEdge)
            os << ".";
          else
            {
              int numberGhostingMe = 0;
              int isAGhost = 0;
              int isMe;
              Node *neighbor;
              
              isMe = (current->myNodeNum == myId);
              mainNeighbors->Reset();
              
              while ((neighbor = mainNeighbors->Next()))
                {
                  if (neighbor->HisGhosts->Member(current))
                    {
                      if (!isMe)
                        cerr << "Other node ghosting something I don't own\n";
                      numberGhostingMe++;
                    }
                  if (neighbor->MyGhosts->Member(current))
                    {
                      if (isAGhost)
                        cerr << "Is a ghost already set\n";
                      isAGhost = 1;
                    }
                }

              if (isAGhost)
                {
                  if ((numberGhostingMe > 0) || (isMe))
                    cerr << "Trouble when isAGhost set\n";
                  os << (char) ('a' + current->myNodeNum);
                }
              else if (numberGhostingMe > 0)
                {
                  os << numberGhostingMe;
                }
              else if (isMe)
                os << "-";
              else
                os << (char) ('A' + current->myNodeNum);
              
               
            }
          
        }
      os << "\n";
    }
  os << "\n";
  
}

void Node::CreateBuffers()
{
  Node *current;
  
  mainNeighbors->Reset();

  while ((current = mainNeighbors->Next()))
    {
      current->outgoingBuffer = (FishData *)
        malloc(sizeof(FishData) * (numFish + 1));
      current->incomingBuffer = (FishData *)
        malloc(sizeof(FishData) * (numFish + 1));
      current->outgoingBuffer[0].fish_num = -1;
      current->incomingBuffer[0].fish_num = -1;
    }
}

/* Stage one communication consists of moving any fish
   from this node to the node who owns the grid that
   the fish has moved to.... basically, pushing fish */
void Node::FillStageOneFor(int node_id)
{
  Node *current = allNodes[node_id];
  FishData *outgoing = current->outgoingBuffer;
  int place = 1;
 
  FishList *outgoingToHim = current->outgoingFish;
  outgoingToHim->Reset();
  Fish *f;
  
  while ((f = outgoingToHim->Next()))
    {
      outgoing[place].px = f->px;
      outgoing[place].py = f->py;
      outgoing[place].vx = f->vx;
      outgoing[place].vy = f->vy;
      outgoing[place].fish_num = f->myId;
      outgoing[place].grid_num = f->grid->myId;
      place++;
    }

  outgoingToHim->RemoveAll();

  outgoing[0].fish_num = place;
}

void Node::ReceiveStageOneFrom(int node_id)
{
  /* just read all the information from the buffer, and place the
     fish appropriately */
  /* we will be getting all the fish he is pushing on us */

  Node *current = allNodes[node_id];
  FishData *incoming = current->incomingBuffer;
  Fish *f;
  GridElement *ge;

  /* first, since we are about to overwrite all grid elements this guy
     ghosts for me, we should clear them out*/

  /* now read off the buffer */
  
  for (int i = 1; i < incoming[0].fish_num; i++)
    {
      FishData *fd = &(incoming[i]);
      f = newFish[fd->fish_num];
      
      f->px = fd->px;
      f->py = fd->py;
      f->vx = fd->vx;
      f->vy = fd->vy;

      ge = theGrid->GridArray()[fd->grid_num];
      ge->AddNewFish(f);
      myFish->Add(f);
    }
  
  incoming[0].fish_num = -1; 
}

void Node::FillAllStageOne()
{
  Node *current;
  mainNeighbors->Reset();
  
  while ((current = mainNeighbors->Next()))
    FillStageOneFor(current->myId);

}

void Node::ReceiveAllStageOne()
{
  Node *current;
  mainNeighbors->Reset();
  
  while ((current = mainNeighbors->Next()))
    ReceiveStageOneFrom(current->myId);

}

/* Stage two consists of sending all of the fish that are in his ghosts
   to him... (i.e., he has a ghostcell that I own, so I have to send him
   the fish that go in it) */

void Node::FillStageTwoFor(int node_id)
{
  Node *current = allNodes[node_id];
  FishData *outgoing = current->outgoingBuffer;
  int place = 1;
  FishList *outgoingToHim = NULL;
  Fish *f;
  
  /* now, look in all of the grid elements that are his ghosts, and
     output all the fish in those elements to him */
  current->HisGhosts->Reset();
  GridElement *ge;

  while ((ge = current->HisGhosts->Next()))
    {
      outgoingToHim = ge->myNewFish;
      outgoingToHim->Reset();

      while ((f = outgoingToHim->Next()))
        {
          outgoing[place].px = f->px;
          outgoing[place].py = f->py;
          outgoing[place].vx = f->vx;
          outgoing[place].vy = f->vy;
          outgoing[place].fish_num = f->myId;
          outgoing[place].grid_num = f->grid->myId;
          place++;
        }
      
    }

  /* remember to write the number of elements in this array into the
     first fish in the buffer */
  
  outgoing[0].fish_num = place;
}

void Node::ReceiveStageTwoFrom(int node_id)
{
  /* just read all the information from the buffer, and place the
     fish appropriately */
  /* we will be getting information on the grids we need ghosted */

  Node *current = allNodes[node_id];
  FishData *incoming = current->incomingBuffer;
  Fish *f;
  GridElement *ge;
  
  /* now read off the buffer */
  
  for (int i = 1; i < incoming[0].fish_num; i++)
    {
      FishData *fd = &(incoming[i]);
      f = newFish[fd->fish_num];
      
      f->px = fd->px;
      f->py = fd->py;
      f->vx = fd->vx;
      f->vy = fd->vy;

      ge = theGrid->GridArray()[fd->grid_num];
      ge->AddNewFish(f);
    }
  incoming[0].fish_num = -1;  
}


void Node::FillAllStageTwo()
{
  Node *current;
  mainNeighbors->Reset();
  
  while ((current = mainNeighbors->Next()))
    FillStageTwoFor(current->myId);

}

void Node::ReceiveAllStageTwo()
{
  Node *current;
  mainNeighbors->Reset();
  
  while ((current = mainNeighbors->Next()))
    ReceiveStageTwoFrom(current->myId);
}

void Node::FillFishInfoBuffer()
{
  /* Load all the fish in my grid element into the FishInfoOutBuffer */
  GridElement *ge;
  Fish *f;
  FishData *fd;
  int place = 1;
  
  myElements->Reset();

  while ((ge = myElements->Next()))
    {
      ge->myCurrentFish->Reset();

      while ((f = ge->myCurrentFish->Next()))
        {
          fd = &(fishInfoOutBuffer[place]);
          fd->fish_num = f->myId;
          fd->grid_num = ge->myId;
          fd->px = f->px;
          fd->py = f->py;
          fd->vx = f->vx;
          fd->vy = f->vy;
          place++;
        }
    }
  fishInfoOutBuffer[0].fish_num = place;
}

void Node::ReceiveFishInfoBuffers()
{
  int node_num;
  FishData *fd;
  Fish *f;
  
  /* For all the incoming fish info buffers, go in and read
     out all the information, updating the proper fish */
  for (node_num = 1; node_num < gNumberNodes; node_num++)
    {

      for (int i = 1; i < fishInfoInBuffer[node_num][0].fish_num; i++)
        {
          fd = &(fishInfoInBuffer[node_num][i]);
          f = currentFish[fd->fish_num];
          f->px = fd->px;
          f->py = fd->py;
          f->vx = fd->vx;
          f->vy = fd->vy;
        }
      fishInfoInBuffer[node_num][0].fish_num = -1;
    }
}

void Node::SetupForCalculations()
{
  /* Conditions true upon entering here...
     the currentFish list represents the Fish that are
     currently in theGrid, and are as they should be..
     the newFish list represents the outdated information that
     needs to be thrown out...  they will be in the GridElements,
     but under the newFish list..

     Things to do in this routine:
     o Clear all the newFish grid elements...
       in both the ones I own, and my Ghosts -- these are the new fish
     o Clear myFish -- these are the current fish
     o Clear outgoin in each node -- these are the current fish...  
     */

  Node *n;
  
  for (int i = 0; i < myNumElementsAndGhosts; i++)
    myElementsAndGhosts[i]->myNewFish->RemoveAll();

  myFish->RemoveAll();

  mainNeighbors->Reset();

  while ((n = mainNeighbors->Next()))
    n->outgoingFish->RemoveAll();
  
}

void Node::FinalizeCalculations()
{
  /* These to do in here
     o Do a switch on all the grid elements... both on myElements
       and all myGhosts
     o switch the main fish list in the node */

  Fish **temp = currentFish;
  for (int i = 0; i < myNumElementsAndGhosts; i++)
    myElementsAndGhosts[i]->SwitchToNew();

  currentFish = newFish;
  newFish = temp;
}


static Fish **surrounding = NULL;

double Node::PerformCalcForGE(double deltaTime, double Time, double &energy,
                              GridElement *ge)
{
  Fish *current, *movingfish;  
  Fish *temp;
  GridElement *dest, *neigh;
#ifdef _TRACKING_
  Fish *tracker[gNumberFish];
  int checking = -1;
#endif
  double maxVel, result;
  int num_surrounding, i;
  
  maxVel = -1.0;
  
#ifdef _TRACKING_
  for (int i = 0; i < gNumberFish; i++)
    tracker[i] = NULL;
#endif

  if (surrounding == NULL)
    surrounding = new (Fish *)[gNumberFish];
  

  /* Just for convenience, we shove all the neighboring fish
     into one array... Note, that this will put the fish in the
     current grid in the array to, so we have to be careful when
     going through it; we need to recognize when we are calcing
     on the same fish */

  num_surrounding = 0;
  
  ge->myCurrentFish->Reset();
  
  while ((current = ge->myCurrentFish->Next()))
    surrounding[num_surrounding++] = current;

  for (i = 0; i < 8; i++)
    {
      neigh = ge->myNeighbors[i];
      neigh->myCurrentFish->Reset();

      while ((current = neigh->myCurrentFish->Next()))
        surrounding[num_surrounding++] = current;
    }

  /* Now, we are ready to calculate all the forces on the fish in
     this grid */

  ge->myCurrentFish->Reset();

  while ((current = ge->myCurrentFish->Next()))
    {
      /* remember to zero out old forces */
      current->fx = 0.0;
      current->fy = 0.0;
      current->potential = 0.0;
      movingfish = newFish[current->myId];

      /* Calculate the force seen by this fish due to all others */
      for (i = 0; i < num_surrounding; i++)
        if (current != surrounding[i])
          {
            temp = surrounding[i];
            current->CalcForce(surrounding[i]);
            if (gCalcEnergy)
              current->CalcPotential(temp);
#ifdef _TRACKING_
            if (current->myId == checking)
              tracker[temp->myId] = temp;
#endif
            
          }
      
      if (gCalcEnergy)
        energy += current->KineticEnergy() + 0.5 * current->potential;

      /* Calculate the new position of the fish, and put the info
         in the movingfish structure */
      result = current->UpdateFish(deltaTime, movingfish);
      maxVel = MAX(maxVel, result);
      
#ifdef _TRACKING_      
      if (current->myId == checking)
        {       
          cout << "Calc'ing Fish #" << current->myId << " at " << Time
               << " on " << ge->myId << "\n";
          cout << "  original: " << current->px << " " << current->py
               << " " << current->vx << " " << current->vy << "\n";
          cout << "  using:\n";
          for (int i = 0; i < gNumberFish; i++)
            if (tracker[i])
              cout << "      Fish #" << i << " " << tracker[i]->px <<
                " " << tracker[i]->py << " " << tracker[i]->vx << " " <<
                tracker[i]->vy << "\n";
          cout << "  update: " << movingfish->px << " " << movingfish->py
               << " " << movingfish->vx << " " << movingfish->vy << "\n";
        }
#endif
      
      /* now check to see if it is still in the same grid element */
      dest = theGrid->GEFromPoint(movingfish->px, movingfish->py);

      movingfish->grid = dest;
      if (dest->isGhost)
        allNodes[dest->myNodeNum]->outgoingFish->Add(movingfish);
      else
        {
          dest->AddNewFish(movingfish);
          myFish->Add(movingfish);
        }
    }

  return maxVel;  
}

double Node::PerformCalcsForList(double deltaTime, double Time, double &energy,
                                 GEList *gridList)
{
  GridElement *current;
  double result, temp;

  result = 0.0;

  for (gridList->Reset(); ((current = gridList->Next()));)
    {
      temp = PerformCalcForGE(deltaTime, Time, energy, current);
      result = MAX(result, temp);
    }
  return result;  
}

double Node::PerformAllCalcs(double deltaTime, double Time, double &energy)
{
  GridElement *current;
  double result, temp;

  result = 0.0;
  
  myElements->Reset();

  while ((current = myElements->Next()))
    {
      temp = PerformCalcForGE(deltaTime, Time, energy, current);
      result = MAX(result, temp);
    }
  return result;
}
  
GridElement **Grid::GridArray()
{
  return myGrid;
}

void Grid::TestPrint(ostream &os)
{
  /* Print out what grid belong to this node */
  /* Print out all this node's neighbors, and which cells
     are shadowed by the neighbor, and which cells this guy
     shadows for the neighbor */
  Node *mainNode = myNodeList[myMainNode];
  GridElement *current;
  
  mainNode->myElements->Reset();
  os << "Grid information print out: Node #" << myMainNode << " of " <<
    myNumberNodes << "nodes\n\n";
  os << "  Owns " << mainNode->myElements->NumElements() << ": ";

  while ((current = mainNode->myElements->Next()))
    current->QuickPrint(os);
  os << "\n";

  for (int i = 0; i < myNumberNodes; i++)
    if (myNodeList[i]->HisGhosts->NumElements() != 0)
      {
        Node *n = myNodeList[i];
        os << "  Neighbor #" << i << ":\n";
        os << "    His ghosts: ";

        n->HisGhosts->Reset();
        n->MyGhosts->Reset();

        while ((current = n->HisGhosts->Next()))
          current->QuickPrint(os);

        os << "\n";
        os << "    My ghosts: ";

        while ((current = n->MyGhosts->Next()))
          current->QuickPrint(os);
        os << "\n";
      }
  os << "\n";
}

void Grid::SetNodeRegion(double left, double top, double right, double bottom,
                         int node_number)
{
  int l = (int) (left * gInvCutoff + 1);
  int r = (int) (right * gInvCutoff + 1);
  int b = (int) (bottom * gInvCutoff + 1);
  int t = (int) (top * gInvCutoff + 1);

  SetNodeRegion(l,t,r,b, node_number);
}

void Grid::SetNodeRegion(int left, int top, int right, int bottom,
                         int node_number)
{
  Node *current_node = myNodeList[node_number];

  //  cout << "Node #" << node_number << " has " << left << " " << top << " "
  //    << right << " " << bottom << "\n";
  
  for (int row = top; row < bottom; row++)
    for (int col = left; col < right; col++)
      {
        GridElement *current = myGrid[(row + 1) * gGridWidth + col + 1];
        current->myNodeNum = node_number;
        current_node->myElements->Add(current);
      }
  
}





