#ifndef STRUCTS_H
#define STRUCTS_H

#include <iostream.h>
#include "List.H"

class GridElement;

class Fish
{
public:
  Fish();
  Fish(double px, double py, double vx, double vy, int id);
  void TestPrint(ostream &os);

  /* these are the base operations in calculating the forces acting
     on a fish */
  void CalcForce(Fish *interactingFish);
  void CalcPotential(Fish *interactingFish);
  double KineticEnergy();
  double UpdateFish(double deltaTime, Fish *movingFish);
  Fish *Clone();
  /* returns the velocity squared */
  
  double px, py;
  double vx, vy;
  double fx, fy;
  double potential;
  int myId;
  GridElement *grid;
};

typedef ListGenerator<Fish> FishListGenerator;
typedef List<Fish> FishList;

class GridElement
{
public:
  GridElement(double left, double right, double top, double bottom,
              int grid_id);
  void TestPrint(ostream &os);
  void QuickPrint(ostream &os);

  int IsNeighboring(int grid_id);
  void SwitchToNew();
  
  double left, right, top, bottom;
  int isEdge;  /* does this GE represent something outside the boundary */
  int isGhost; /* does this represent a GE for another node */
  int myNodeNum; /* which node actually controls this grid element */
  GridElement **myNeighbors;
  int myId;

  FishList *myCurrentFish;
  FishList *myNewFish;
  
  void AddNewFish(Fish *);
  void AddCurrentFish(Fish *);
  
  void RemoveNewFish(Fish *);
  void RemoveCurrentFish(Fish *);
  
};

typedef struct
{
  int fish_num;
  int grid_num;
  double px, py;
  double vx, vy;
} FishData;

typedef ListGenerator<GridElement> GEListGenerator;
typedef List<GridElement> GEList;

/* Nodes represent all the information for a processor.  There are two
   types:  The main node, and other nodes.  The main node is the node
   that represents this processor's responsibilities.. The other nodes
   are this processor's representations for all the other processor's
   in the network */

class Node;
class Grid;
typedef ListGenerator<Node> NodeListGenerator;
typedef List<Node> NodeList;

class Node
{
public:
  /* only use the next constructor when you are creating nodes that are
     not the main node for this processor */
  Node(int node_id);
  /* only use the next constructor when you are creating the main node */
  /* it will allocate all the memory for the fish, the neighboring nodes,
     and the grid.. It also does the grid layout... a real work horse */
  Node(int node_id, int number_nodes, int number_fish, Fish **the_fish); 
  void TestPrint(ostream &os);

  /* These methods are only used when this node is the main node */
  void CreateBuffers();
  void FillStageOneFor(int node_id);
  void ReceiveStageOneFrom(int node_id);
  void FillAllStageOne();
  void ReceiveAllStageOne();

  void FillStageTwoFor(int node_id);
  void ReceiveStageTwoFrom(int node_id);
  void FillAllStageTwo();
  void ReceiveAllStageTwo();
  
  void FillFishInfoBuffer();
  void ReceiveFishInfoBuffers();
  
  /* New style of Performing calcs */
  void SetupForCalculations();
  double PerformCalcForGE(double deltaTime, double Time, double &energy,
                          GridElement *ge);
  double PerformAllCalcs(double deltaTime, double Time, double &energy);
  double PerformCalcsForList(double deltaTime, double Time, double &energy,
                            GEList *gridlist);
  void FinalizeCalculations();

  void ComputeGridDependencies();

  void FillDependencies(Grid *theGrid, GEList *myElements);
  
  GEList *myElements;

  /* The following are for neighbor nodes */
  GEList *HisGhosts;  /* these are the grid elements that this node
                         is responsible for, that this neighbor is
                         shadowing */
  GEList *MyGhosts;
  GEList *myStage1Dependencies;
  GEList *myStage2Dependencies;
  
  /* back to the main node fields */
  GEList *myStage1Leftovers;
  GEList *myStage2Leftovers;
  Node **myStage1Ordering;
  Node **myStage2Ordering;
  GridElement **myElementsAndGhosts;
  int myNumElementsAndGhosts;
  
  /* this always has new fish in it */
  FishList *outgoingFish; /* at each iteration, this node will place fish
                             on this list if it sees that fish moves into
                             this neighbors territory.. */
  int myId;

  /* For the interlaced processing */
  
  /* for debugging purposes */
  void GraphLayout(ostream &os);
  void GraphNeighbors(ostream &os);

  /* These fields are only in use when this node object happens
     to be the main node for this processor */
  Node **allNodes;
  Fish **newFish;
  Fish **currentFish;
  int numFish;
  int numNodes;
  NodeList *mainNeighbors;
  Grid *theGrid;
  FishData *fishInfoOutBuffer; /* buffer used to communicated with node 0.
                                  this will broadcast all this node's fish
                                  data for display */
  FishData **fishInfoInBuffer; /* if this is node zero, these are the
                                  receive buffers from all the other nodes */
  FishList *myFish;
  /* These are fields only for neighbors of the mainNode */
  FishData *incomingBuffer;
  FishData *outgoingBuffer;
  
};


class Grid
{
public:
  Grid(Node **node_list, FishListGenerator *fgen, FishListGenerator *curgen,
       int number);
  ~Grid();

  void PartitionGrid(int number); /* divide the grid up into number parts */
  NodeList *ComputeNeighborInfo(int myNodeNum); /* given that the grid
                                                   elements are labeled with
                                                   node numbers, figure out
                                                   who needs to shadow who */
  GridElement *GEFromPoint(double x, double y);
  GEListGenerator *CreateGenerator();
  
  /* for debugging purposes */
  int VerifyLayout();
  void TestPrint(ostream &os);
  GridElement **GridArray();
protected:
  /* this is a helper procedure that marks all the grid elements in a
     region as belonging to a particular node */
  void SetNodeRegion(double left, double top, double right, double bottom,
                     int node_number);
  void SetNodeRegion(int left, int top, int right, int bottom,
                     int node_number);
  void ComputeRegions(int number_rows, int *cols_in_each_row);
  
  Node **myNodeList;
  int myNumberNodes;
  GridElement **myGrid;
  int myMainNode; /* this is the node number for which this grid views as
                     the main node (all others are neighbors) */
};

#endif






