#include "Structures.H"
#include "FishUtils.H"
#include <iostream.h>
#include <fstream.h>
#include "dmartin/tablet.h"
#include <math.h>
#include "mpi.h"
#include "dmartin/timer.h"
#include "List.C"

void Build_MPI_Fish(MPI_Datatype *datatype);
void CollectFishInfo(Node *mainNode, MPI_Datatype &MPI_Fish);
void TieBuffersTogether(Node **nodelist);
Node **CreateAllNodes();
void CompareBuffers(Node *all_node, Node *main_node);
double ProcessAll(Node **mainNodeList, double time, double timeDelta);
void CompareFish(Node *allNode, Node *mainNode);

int main(int argc, char **argv)
{
  MPI_Datatype MPI_Fish;
  MPI_Status status;
  Fish **mainFishList;
  Node **mainNodeList;

  Node *mainNode;
  ofstream tracefile;
  int ourTag = 30;
  double time;
  double timeDelta;
  double maxSqrVel, myMaxSqrVel;
  double nextTraceTime = gTraceInterval;
  double energy;
  Ctablet *window;
  Node *currentNeighbor;
  Ctimer total_time;
  Ctimer comm_time;
  Ctimer calc_time;
  
  int i;

  /* Initialize MPI, and all that */
  MPI_Init (&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &g_i_id);
  MPI_Comm_size(MPI_COMM_WORLD, &gNumberNodes);

  if (!ProcessArguments(argc, argv))
    return -1;

  /* Tell MPI about my FishData structure */
  Build_MPI_Fish(&MPI_Fish);

  /* Configure outselves for the given domain size and cutoff */
  ComputeDimensions();

  //  if (g_i_id == 0)
  //  tracefile.open(gTraceFileName);
  
  /* create the main node .. must pass into it a list of all the fish,
     after they have been initialized with the global allocation feature */

  
  mainNodeList = CreateAllNodes();
  
  mainFishList = new (Fish*)[gNumberFish];
  InitializeFish(mainFishList, gNumberFish);
  mainNode = new Node(g_i_id, gNumberNodes, gNumberFish, mainFishList);
  
  if (gDisplay && (g_i_id == 0))
    {
      window = new  Ctablet();
      window->init("Fish!", gWindowSize, gWindowSize);
    }

  if (g_i_id == 0)
    PrintFishInfo(tracefile, 0.0, mainFishList, 0.0);

  total_time.reset();
  comm_time.reset();
  calc_time.reset();

  total_time.start();
  
  /* Now do the darn simulation */
  for (timeDelta = gDtMin, time = 0.0; time < gEndTime; time += timeDelta)
    {
      maxSqrVel = -1.0;
      energy = 0.0;
      double allNodeMax;
      
      calc_time.start();
      myMaxSqrVel = mainNode->PerformCalculations(timeDelta, time,
                                                  energy);
      calc_time.stop();


      allNodeMax = ProcessAll(mainNodeList, time, timeDelta);
      
      comm_time.start();
      /* Now handle the communications */
      mainNode->FillAllBuffers();

      mainNode->mainNeighbors->Reset();

      
      while ((currentNeighbor = mainNode->mainNeighbors->Next()))
        {
          int count;
          //mainNode->FillBufferFor(currentNeighbor->myId);

          count = (int) currentNeighbor->outgoingBuffer[0].fish_num;
          if (g_i_id < currentNeighbor->myId)
            {
              /* If my rank is lower than his, then I send first,
                 he receives, then sends to me, and I receive */
              MPI_Send(currentNeighbor->outgoingBuffer, count + 1,
                       MPI_Fish, currentNeighbor->myId, ourTag,
                       MPI_COMM_WORLD);
              //cout << "Sent ghosts to " << currentNeighbor->myId << "\n";
              MPI_Recv(currentNeighbor->incomingBuffer, gNumberFish + 1,
                       MPI_Fish, currentNeighbor->myId, ourTag,
                       MPI_COMM_WORLD, &status);

              //cout << "Got ghosts from " << currentNeighbor->myId << "\n";
            }
          else
            {
              /* else, it's the other way around */
              MPI_Recv(currentNeighbor->incomingBuffer, gNumberFish + 1,
                       MPI_Fish, currentNeighbor->myId, ourTag,
                       MPI_COMM_WORLD, &status);
              //cout << "Got ghosts from " << currentNeighbor->myId << "\n";
              MPI_Send(currentNeighbor->outgoingBuffer, count + 1,
                       MPI_Fish, currentNeighbor->myId, ourTag,
                       MPI_COMM_WORLD);
              //cout << "Sent ghosts to " << currentNeighbor->myId << "\n";
            }
          MPI_Get_count(&status, MPI_Fish, &count);
          //          if (count != currentNeighbor->incomingBuffer[0].fish_num)
          //  cout << "We got " << count << " when expecting " <<
          //    currentNeighbor->incomingBuffer[0].fish_num << "\n";
          
          //mainNode->ReceiveBufferFrom(currentNeighbor->myId);
        }
      mainNode->ReceiveAllBuffers();
      
      comm_time.stop();

      /* Here's the kicker... Let's see if our incoming buffers match the
         incoming buffers of the mainList */

      CompareBuffers(mainNodeList[g_i_id], mainNode);

      int total = 0;
      for (int y = 0; y < gNumberNodes; y++)
        total += mainNodeList[y]->myFish->NumElements();

      if (total != gNumberFish)
        cout << "We only have " << total << " fish\n";
      
      if (gDisplay)
        {
          comm_time.start();
          /* Collect all the fish information */
          CollectFishInfo(mainNode, MPI_Fish);
          comm_time.stop();
          
          /* if the trace interval time was been tripped, print it out */
          if ((time > nextTraceTime) && (g_i_id == 0))
            {
              PrintFishInfo(tracefile, time, mainFishList, energy);
              nextTraceTime += gTraceInterval;
            }
          
          /* Draw all the fish */
          if (g_i_id == 0)
            {
              for (i = 0; i < gNumberFish; i++)
                DrawFish(window, mainFishList[i]);
              
              window->renderScene();    
            }
        }

      comm_time.start();
      /* now get the darn reduced maxSqrVel from all the other nodes */
      MPI_Allreduce (&myMaxSqrVel, &maxSqrVel, 1, MPI_DOUBLE, MPI_MAX,
                     MPI_COMM_WORLD);
      comm_time.stop();

      if (allNodeMax != maxSqrVel)
        cout << "Difference on the vel\n";
      
      /* make sure to update the timeDelta */
      timeDelta = gDxMin / sqrt(maxSqrVel);
      timeDelta = MIN(gDtMax, timeDelta);
      timeDelta = MAX(gDtMin, timeDelta);
    }

  /* Just for the record, we get the last state of the fish */

  total_time.stop();
  CompareBuffers(mainNodeList[g_i_id], mainNode);


  CollectFishInfo(mainNode, MPI_Fish);
  for (i = 1; i < gNumberNodes; i++)
    mainNodeList[i]->FillFishInfoBuffer();
  
  mainNodeList[0]->ReceiveFishInfoBuffers();

  if (g_i_id == 0)
    {
      CompareFish(mainNodeList[g_i_id], mainNode);
      
      tracefile.open("traceA1");
      PrintFishInfo(tracefile, time, mainFishList, energy);
      tracefile.close();

      tracefile.open("traceB1");
      PrintFishInfo(tracefile, time, mainNodeList[0]->allFish, energy);
      tracefile.close();
      
      cout << "\n";
      cout << "Total time: " << total_time.value() << "\n";
      cout << " Comm time: " << comm_time.value() << "\n";
      cout << " Calc time: " << calc_time.value() << "\n";
      cout << "\n";
    }
  
  
  MPI_Finalize();

  return 0;
}

void Build_MPI_Fish(MPI_Datatype *datatype)
{
  FishData *sample;
  int block_lengths[6];
  MPI_Aint displacements[6];
  MPI_Aint addresses[7];
  MPI_Datatype typelist[6];

  typelist[0] = MPI_INT;
  typelist[1] = MPI_INT;
  typelist[2] = MPI_DOUBLE;
  typelist[3] = MPI_DOUBLE;
  typelist[4] = MPI_DOUBLE;
  typelist[5] = MPI_DOUBLE;

  block_lengths[0] = 1;
  block_lengths[1] = 1;
  block_lengths[2] = 1;
  block_lengths[3] = 1;
  block_lengths[4] = 1;
  block_lengths[5] = 1;

  sample = (FishData *) malloc(sizeof(FishData));

  MPI_Address(sample, &addresses[0]);
  MPI_Address(&(sample->fish_num), &addresses[1]);
  MPI_Address(&(sample->grid_num), &addresses[2]);
  MPI_Address(&(sample->px), &addresses[3]);
  MPI_Address(&(sample->py), &addresses[4]);
  MPI_Address(&(sample->vx), &addresses[5]);
  MPI_Address(&(sample->vy), &addresses[6]);
  
  displacements[0] = addresses[1] - addresses[0];
  displacements[1] = addresses[2] - addresses[0];
  displacements[2] = addresses[3] - addresses[0];
  displacements[3] = addresses[4] - addresses[0];
  displacements[4] = addresses[5] - addresses[0];
  displacements[5] = addresses[6] - addresses[0];

  MPI_Type_struct(6, block_lengths, displacements, typelist,
                  datatype);

  MPI_Type_commit(datatype);

}


void CollectFishInfo(Node *mainNode, MPI_Datatype &MPI_Fish)
{
  int i;
  int ourTag2 = 33;
  MPI_Status status;
  
  if (g_i_id == 0)
    {
      /* then we must receive the information from all the other
         nodes */
      for (i = 1; i < gNumberNodes; i++)
        {
          MPI_Recv(mainNode->fishInfoInBuffer[i], gNumberFish + 1,
                   MPI_Fish, i, ourTag2, MPI_COMM_WORLD, &status);
          //cout << "Got fish from " << i << "\n";
        }
      mainNode->ReceiveFishInfoBuffers();
    }
  else
    {
      /* we must send our porition of the fish to the main node */
      mainNode->FillFishInfoBuffer();
      MPI_Send(mainNode->fishInfoOutBuffer,
               (int) mainNode->fishInfoOutBuffer[0].fish_num,
               MPI_Fish, 0, ourTag2, MPI_COMM_WORLD);
      //cout << "Sent fish to " << 0 << "\n";
    }
}



Node **CreateAllNodes()
{
  Node **mainNodeList;
  Fish **mainFishList;
  int i;
  
  mainNodeList = new (Node*)[gNumberNodes];
  
  /* create the main node list... Each of these nodes believes it
     is on a single processor.. we can trick them so that they will
     communicate to each other eventhough they are on the same processor */

  for (i = 0; i < gNumberNodes; i++)
    {
      mainFishList = new (Fish*)[gNumberFish];
      InitializeFish(mainFishList, gNumberFish);
      mainNodeList[i] = new Node(i, gNumberNodes, gNumberFish, mainFishList);
    }
  
  /* Now, we have to patch up their communications buffers so that
     they automatically talk to one another... This is a little of a
     hack, but a decent one */
  TieBuffersTogether(mainNodeList);

  return mainNodeList;
}

double ProcessAll(Node **mainNodeList, double time, double timeDelta)
{
  double temp;
  double maxSqrVel = 0.0;
  int i;
  double energy;

  energy = 0.0;
  
  for (i = 0; i < gNumberNodes; i++)
    {
      temp = mainNodeList[i]->PerformCalculations(timeDelta, time,
                                                         energy);
      maxSqrVel = MAX(temp, maxSqrVel);
      
      mainNodeList[i]->FillAllBuffers();
    }
  
  for (i = 0; i < gNumberNodes; i++)
    mainNodeList[i]->ReceiveAllBuffers();

  return maxSqrVel;
}


void TieBuffersTogether(Node **nodelist)
{

  /* first, fix the fishInfo buffers so that all the ones in Node 0 are
     shared appropriately */
  int i;
  
  for (i = 1; i < gNumberNodes; i++)
    {
      FishData *temp = nodelist[i]->fishInfoOutBuffer;
      /* make the out from node i be the same as in the buffer from node i
         on node 0 */
      nodelist[i]->fishInfoOutBuffer = nodelist[0]->fishInfoInBuffer[i];
      free(temp);
    }


  /* now, fix the neighboring buffers that govern normal inter node
     communication */

  /* basically, put all the outgoing buffers from node a to be node b,
     as the incoming buffers from node b to node a... */
  for (i = 0; i < gNumberNodes; i++)
    {
      Node *source = nodelist[i];
      Node *currentNeighbor;
      Node *dest, *sour;
      
      source->mainNeighbors->Reset();

      while ((currentNeighbor = source->mainNeighbors->Next()))
        {
          FishData *temp;
          dest = currentNeighbor;
          sour = nodelist[currentNeighbor->myId]->allNodes[source->myId];

          temp = dest->outgoingBuffer;
          dest->outgoingBuffer = sour->incomingBuffer;

          free(temp);
        }
      
    }
}

void CompareBuffers(Node *all_node, Node *main_node)
{
  Node *all_neighbor;
  Node *main_neighbor;
  FishData *adata;
  FishData *mdata;
  Fish *mcurrent;
  Fish *acurrent;
  
  all_node->myFish->Reset();
  main_node->myFish->Reset();

  while ((mcurrent = main_node->myFish->Next()))
    {
      acurrent = all_node->myFish->Next();

      if (acurrent->myId != mcurrent->myId)
        cout << "Mismatch on fish ids\n";
      if (acurrent->px != mcurrent->px)
        cout << "Mistmatch in fish px\n";
      if (acurrent->py != mcurrent->py)
        cout << "Mistmatch in fish py\n";
      if (acurrent->vx != mcurrent->vx)
        cout << "Mistmatch in fish vx\n";
      if (acurrent->vy != mcurrent->vy)
        cout << "Mistmatch in fish vy\n";
    }
  
  all_node->mainNeighbors->Reset();
  main_node->mainNeighbors->Reset();

  while ((all_neighbor = all_node->mainNeighbors->Next()))
    {
      main_neighbor = main_node->mainNeighbors->Next();

      if (main_neighbor->myId != all_neighbor->myId)
        cout << "Oh oh... mismatch on the ids\n";

      if (main_neighbor->incomingBuffer[0].fish_num !=
          all_neighbor->incomingBuffer[0].fish_num)
        {
          cout << "Hell.. mismatch on amount of data sending: \n";
          cout << "  Main: " << main_neighbor->incomingBuffer[0].fish_num
               << "\n";
          cout << "  All: " << all_neighbor->incomingBuffer[0].fish_num
               << "\n\n";
        }
      
      adata = all_neighbor->incomingBuffer;
      mdata = main_neighbor->incomingBuffer;
      for (int i = 1; i < adata[0].fish_num; i++) 
        {
          if (adata[i].fish_num != mdata[i].fish_num)
            cout << "Fish num difference\n";
          if (adata[i].grid_num != mdata[i].grid_num)
            cout << "Grid num difference\n";
          if (adata[i].px != mdata[i].px)
            cout << "Px difference\n";
          if (adata[i].py != mdata[i].py)
            cout << "Py difference\n";
          if (adata[i].vx != mdata[i].vx)
            cout << "Vx difference\n";
          if (adata[i].vy != mdata[i].vy)
            cout << "Vy difference\n";
              
              
        }
    }
}

void CompareFish(Node *allNode, Node *mainNode)
{
  Fish *acurrent, *mcurrent;
  
  for (int i = 0; i < gNumberFish; i++)
    {
      acurrent = allNode->allFish[i];
      mcurrent = mainNode->allFish[i];

      if (acurrent->myId != mcurrent->myId)
        cout << "Mismatch on fish ids " << acurrent->myId << "\n";
      if (acurrent->px != mcurrent->px)
        cout << "Mistmatch in fish px " << acurrent->myId << "\n";
      if (acurrent->py != mcurrent->py)
        cout << "Mistmatch in fish py " << acurrent->myId << "\n";
      if (acurrent->vx != mcurrent->vx)
        cout << "Mistmatch in fish vx " << acurrent->myId << "\n";
      if (acurrent->vy != mcurrent->vy)
        cout << "Mistmatch in fish vy " << acurrent->myId << "\n";
    }
  
}
