#include "Structures.H"
#include "FishUtils.H"
#include <iostream.h>
#include <fstream.h>
#include "dmartin/tablet.h"
#include <math.h>

void TieBuffersTogether(Node **nodelist);
double PerformStage1(Node *n, double timeDelta, double time, double &energy);
double PerformStage2(Node *n, double timeDelta, double time, double &energy);
double PerformLeftover(Node *n, double timeDelta, double time, double &energy);

int main(int argc, char **argv)
{
  Fish **mainFishList;
  Node **mainNodeList;
  ofstream tracefile;
  double time;
  double timeDelta;
  double maxSqrVel;
  double nextTraceTime = gTraceInterval;
  double energy;
  Ctablet *window;
  
  int i;
  if (!ProcessArguments(argc, argv))
    return -1;

  ComputeDimensions();
  
  mainNodeList = new (Node*)[gNumberNodes];
  
  tracefile.open(gTraceFileName);
  
  /* create the main node list... Each of these nodes believes it
     is on a single processor.. we can trick them so that they will
     communicate to each other eventhough they are on the same processor */

  for (i = 0; i < gNumberNodes; i++)
    {
      mainFishList = new (Fish*)[gNumberFish];
      InitializeFish(mainFishList, gNumberFish);
      mainNodeList[i] = new Node(i, gNumberNodes, gNumberFish, mainFishList);
    }
  
  /* Now, we have to patch up their communications buffers so that
     they automatically talk to one another... This is a little of a
     hack, but a decent one */
  TieBuffersTogether(mainNodeList);

  if (gDisplay)
    {
      window = new  Ctablet();
      window->init("Fish!", gWindowSize, gWindowSize);
    }

  PrintFishInfo(tracefile, 0.0, mainNodeList[0]->currentFish, 0.0);
  
  for (timeDelta = gDtMin, time = 0.0; time < gEndTime; time += timeDelta)
    {
      maxSqrVel = -1.0;
      energy = 0.0;
      
      for (i = 0; i < gNumberNodes; i++)
        {
          double temp;
          mainNodeList[i]->SetupForCalculations();
          temp = PerformStage1(mainNodeList[i], timeDelta, time, energy);
          maxSqrVel = MAX(temp, maxSqrVel);
              
          mainNodeList[i]->FillAllStageOne();
        }

      for (i = 0; i < gNumberNodes; i++)
        mainNodeList[i]->ReceiveAllStageOne();

      for (i = 0; i < gNumberNodes; i++)
        {
          double temp;
          temp = PerformStage2(mainNodeList[i], timeDelta, time, energy);
          maxSqrVel = MAX(temp, maxSqrVel);
              
          mainNodeList[i]->FillAllStageTwo();
        }

      for (i = 0; i < gNumberNodes; i++)
        mainNodeList[i]->ReceiveAllStageTwo();

      for (i = 0; i < gNumberNodes; i++)
        {
          double temp;
          temp = PerformLeftover(mainNodeList[i], timeDelta, time, energy);
          maxSqrVel = MAX(temp, maxSqrVel);

          mainNodeList[i]->FinalizeCalculations();
        }
      
      if (gDisplay)
        {
          /* Collect all the fish information */
          for (i = 1; i < gNumberNodes; i++)
            mainNodeList[i]->FillFishInfoBuffer();
          
          mainNodeList[0]->ReceiveFishInfoBuffers();
          
          /* if the trace interval time was been tripped, print it out */
          if (time > nextTraceTime)
            {
              PrintFishInfo(tracefile, time, mainNodeList[0]->currentFish,
                            energy);
              nextTraceTime += gTraceInterval;
            }
          
          /* Draw all the fish */
          if (gDisplay)
            {
              for (i = 0; i < gNumberFish; i++)
                DrawFish(window, mainNodeList[0]->currentFish[i]);
              
              window->renderScene();    
            }
        }

      /* make sure to update the timeDelta */
      timeDelta = gDxMin / sqrt(maxSqrVel);
      timeDelta = MIN(gDtMax, timeDelta);
      timeDelta = MAX(gDtMin, timeDelta);
    }

  /* one final time, get fish info */
  for (i = 1; i < gNumberNodes; i++)
    mainNodeList[i]->FillFishInfoBuffer();
  
  mainNodeList[0]->ReceiveFishInfoBuffers();
  PrintFishInfo(tracefile, time, mainNodeList[0]->currentFish, energy);
  
  tracefile.close();
}

void TieBuffersTogether(Node **nodelist)
{

  /* first, fix the fishInfo buffers so that all the ones in Node 0 are
     shared appropriately */
  int i;
  
  for (i = 1; i < gNumberNodes; i++)
    {
      FishData *temp = nodelist[i]->fishInfoOutBuffer;
      /* make the out from node i be the same as in the buffer from node i
         on node 0 */
      nodelist[i]->fishInfoOutBuffer = nodelist[0]->fishInfoInBuffer[i];
      free(temp);
    }


  /* now, fix the neighboring buffers that govern normal inter node
     communication */

  /* basically, put all the outgoing buffers from node a to be node b,
     as the incoming buffers from node b to node a... */
  for (i = 0; i < gNumberNodes; i++)
    {
      Node *source = nodelist[i];
      Node *currentNeighbor;
      Node *dest, *sour;
      
      source->mainNeighbors->Reset();

      while ((currentNeighbor = source->mainNeighbors->Next()))
        {
          FishData *temp;
          dest = currentNeighbor;
          sour = nodelist[currentNeighbor->myId]->allNodes[source->myId];

          temp = dest->outgoingBuffer;
          dest->outgoingBuffer = sour->incomingBuffer;

          free(temp);
        }
      
    }
}

double PerformStage1(Node *n, double timeDelta, double time, double &energy)
{
  double result = 0.0;
  double temp;
  Node **ordering = n->myStage1Ordering;
  
  for (int i = 0; i < n->mainNeighbors->NumElements(); i++)
    {
      temp = n->PerformCalcsForList(timeDelta, time, energy,
                                    ordering[i]->myStage1Dependencies);
      result = MAX(result, temp);
    }
  return result;
}

double PerformStage2(Node *n, double timeDelta, double time, double &energy)
{
  double result = 0.0;
  double temp;
  Node **ordering = n->myStage2Ordering;
  
  for (int i = 0; i < n->mainNeighbors->NumElements(); i++)
    {
      temp = n->PerformCalcsForList(timeDelta, time, energy,
                                    ordering[i]->myStage2Dependencies);
      result = MAX(result, temp);
    }
  return result;
}

double PerformLeftover(Node *n, double timeDelta, double time, double &energy)
{
  return n->PerformCalcsForList(timeDelta, time, energy,
                                n->myStageLeftovers);
}



